class GumoBaseError(RuntimeError):
    pass


class ConfigurationError(GumoBaseError):
    pass


class ServiceAccountConfigurationError(ConfigurationError):
    pass


class ObjectNotoFoundError(GumoBaseError):
    pass



import dataclasses
import enum

from typing import Optional


@dataclasses.dataclass(frozen=True)
class GoogleCloudProjectID:
    value: str

    def __post_init__(self):
        if not isinstance(self.value, str):
            raise ValueError(f'project_id must be a string, expect: {type(self.value)}')


@dataclasses.dataclass(frozen=True)
class GoogleCloudLocation:
    # TODO: リージョン一覧の取得は、 Google Compute Engine の API を利用して取得できるようにしたい
    #
    # Python 用の Google API Client Library を利用し、Compute Engine API の リージョン一覧
    # メソッドを使用することより、現在利用可能なリージョン一覧を取得できます。
    #
    # Google API Client Library for Python については、
    # 以下のリンク： [1] をご参照ください。
    #
    # また、Compute Engine API のリージョン一覧メソッドにつきましては、
    # 以下のリンク ：[2] をご参照ください。
    #
    # [1] https://developers.google.com/api-client-library/python/apis/compute/v1
    # [2] https://developers.google.com/resources/api-libraries/documentation/compute/v1/python/latest/compute_v1.regions.html#list # noqa: E501
    AVAILABLE_LOCATIONS = [
        'us-west1',  # オレゴン
        'us-west2',  # ロサンゼルス
        'us-central1',  # アイオワ
        'us-east1',  # サウスカロライナ
        'us-east4',  # 北バージニア
        'northamerica-northeast1',  # モントリオール
        'southamerica-east1',  # サンパウロ

        'europe-west1',  # ベルギー
        'europe-west2',  # ロンドン
        'europe-west3',  # フランクフルト
        'europe-west4',  # オランダ
        'europe-north1',  # フィンランド

        'asia-south1',  # ムンバイ
        'asia-southeast1',  # シンガポール
        'asia-east1',  # 台湾
        'asia-east2',  # 香港
        'asia-northeast1',  # 東京
        'australia-southeast1',  # シドニー
    ]
    value: str

    def __post_init__(self):
        if not isinstance(self.value, str):
            raise ValueError(f'location must be a string, expect: {type(self.value)}')

        if self.value not in self.AVAILABLE_LOCATIONS:
            raise ValueError(f'Invalid value of {self.value} is not available locations.')


class ApplicationPlatform(enum.Enum):
    Local = 'local'
    GoogleAppEngine = 'google-app-engine'


@dataclasses.dataclass(frozen=True)
class ServiceAccountCredentialConfig:
    enabled: bool
    bucket_name: Optional[str] = None
    blob_path: Optional[str] = None

    def __post_init__(self):
        if not self.enabled:
            return

        if not isinstance(self.bucket_name, str):
            raise ValueError(f'bucket_name must be a string, expect: {type(self.bucket_name)}')

        if not isinstance(self.blob_path, str):
            raise ValueError(f'blob_path must be a string, expect: {type(self.blob_path)}')


# will be deprecated.
@dataclasses.dataclass(frozen=True)
class ServiceAccountCredentialPath:
    value: Optional[str] = None

    def __post_init__(self):
        if isinstance(self.value, str) and not self.value.startswith('gs://'):
            raise ValueError(f'{self.__class__.__name__} must be starts with gs://')

    def credential_config(self) -> ServiceAccountCredentialConfig:
        if not isinstance(self.value, str):
            return ServiceAccountCredentialConfig(
                enabled=False
            )

        _, _, bucket_name, blob_path = self.value.split('/', 3)
        return ServiceAccountCredentialConfig(
            enabled=True,
            bucket_name=bucket_name,
            blob_path=blob_path,
        )


@dataclasses.dataclass(frozen=True)
class GumoConfiguration:
    google_cloud_project: GoogleCloudProjectID
    google_cloud_location: GoogleCloudLocation
    application_platform: ApplicationPlatform
    service_account_credential_config: ServiceAccountCredentialConfig  # will be deprecated

    @property
    def is_local(self) -> bool:
        return self.application_platform == self.application_platform.Local

    @property
    def is_google_app_engine(self) -> bool:
        return self.application_platform == self.application_platform.GoogleAppEngine



import setuptools


name = 'gumo-core'
version = '0.0.23'
description = 'Gumo Core Library'
dependencies = [
    'pyyaml >= 5.1',
    'injector >= 0.13.1',
    'google-cloud-storage >= 1.13.0',
    'google-api-python-client >= 1.7.4'
]

with open("README.md", "r") as fh:
    long_description = fh.read()

packages = [
    package for package in setuptools.find_packages()
    if package.startswith('gumo')
]

setuptools.setup(
    name=name,
    version=version,
    author="Gumo Project Team",
    author_email="gumo-py@googlegroups.com",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gumo-py/gumo-core",
    packages=packages,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=dependencies,
)



import os
import json

from typing import Union

from google.oauth2 import service_account
from google.auth import compute_engine

from google.cloud import storage
from injector import inject

from gumo.core.exceptions import ServiceAccountConfigurationError
from gumo.core.injector import injector
from gumo.core.domain.configuration import GumoConfiguration


Credentials = Union[service_account.Credentials, compute_engine.Credentials]


class GoogleOAuthCredentialManager:
    _credential = None

    @classmethod
    def get_credential(cls) -> Credentials:
        if cls._credential:
            return cls._credential

        cls._credential = injector.get(cls).build_credential()
        return cls._credential

    @inject
    def __init__(
            self,
            gumo_configuration: GumoConfiguration,
    ):
        self._gumo_configuration = gumo_configuration
        self._credential_config = self._gumo_configuration.service_account_credential_config

    def build_credential(self) -> Credentials:
        _credentials = None

        try:
            if self._gumo_configuration.is_google_app_engine:
                _credentials = compute_engine.Credentials()
            elif 'GOOGLE_APPLICATION_CREDENTIALS' in os.environ:
                _credentials = service_account.Credentials.from_service_account_info(
                    info=self._get_content_from_local()
                )
            elif self._credential_config.enabled:
                # will be deprecated.
                _credentials = service_account.Credentials.from_service_account_info(
                    info=self._get_content_from_storage()
                )
        except ServiceAccountConfigurationError:
            raise
        except RuntimeError as e:
            raise ServiceAccountConfigurationError(e)

        if _credentials is None:
            raise ServiceAccountConfigurationError(f'ServiceAccount Credential Config disabled.')

        return _credentials

    def _get_content_from_storage(self):
        storage_client = storage.Client()
        bucket = storage_client.get_bucket(bucket_name=self._credential_config.bucket_name)
        blob = bucket.blob(blob_name=self._credential_config.blob_path)
        content = blob.download_as_string(client=storage_client)

        return json.loads(content)

    def _get_content_from_local(self):
        credential_path = os.environ.get('GOOGLE_APPLICATION_CREDENTIALS')

        if not os.path.exists(credential_path):
            raise ServiceAccountConfigurationError(f'GOOGLE_APPLICATION_CREDENTIALS={credential_path} is not found.')

        with open(credential_path, 'r') as f:
            content = f.read()

        return json.loads(content)


def get_google_oauth_credential() -> Credentials:
    return GoogleOAuthCredentialManager.get_credential()



import dataclasses
from typing import List
from typing import Union
from typing import Optional

import copy
import base64
import uuid


@dataclasses.dataclass(frozen=True)
class KeyPair:
    kind: str
    name: str


class _BaseKey:
    def parent(self):
        raise NotImplementedError()

    def pairs(self) -> List:
        return []

    def flat_pairs(self) -> List:
        return []

    def kind(self) -> str:
        raise NotImplementedError()

    def name(self) -> str:
        raise NotImplementedError()

    def key_literal(self) -> str:
        raise NotImplementedError()

    def key_path(self) -> str:
        raise NotImplementedError()

    def key_path_urlsafe(self) -> str:
        raise NotImplementedError()


class NoneKey(_BaseKey):
    _root = None

    @classmethod
    def get_instance(cls):
        if cls._root:
            return cls._root
        cls._root = cls()
        return cls._root

    def parent(self):
        return self

    def pairs(self):
        return []

    def flat_pairs(self):
        return []

    def kind(self):
        return None

    def name(self):
        return None

    def __repr__(self):
        return "NoneKey(kind='None', name='none')"

    def __eq__(self, other):
        return isinstance(other, self.__class__)

    def key_literal(self):
        return None

    def key_path(self):
        return None

    def key_path_urlsafe(self):
        return None


@dataclasses.dataclass(frozen=True)
class EntityKey(_BaseKey):
    _pairs: List[KeyPair]

    def __post_init__(self):
        if len(self._pairs) == 0:
            raise ValueError('EntityKey#pairs must have at least one element.')

    def parent(self):
        if len(self._pairs) == 1:
            return NoneKey.get_instance()

        return EntityKey(self._pairs[0:-1])

    def pairs(self):
        return self._pairs

    def flat_pairs(self):
        flat_pairs = []
        for pair in self.pairs():
            flat_pairs.append(pair.kind)
            flat_pairs.append(pair.name)

        return flat_pairs

    def kind(self):
        return self._pairs[-1].kind

    def name(self):
        return self._pairs[-1].name

    def key_literal(self) -> str:
        return 'Key({})'.format(', '.join([
            f"'{i}'" for i in self.flat_pairs()
        ]))

    def key_path(self) -> str:
        """
        Returns a structured key separated by `/`
        The kind and name are separated by `:`
        """
        pairs = []
        for pair in self.pairs():
            pairs.append(f'{pair.kind}:{pair.name}')
        return '/'.join(pairs)

    def key_url(self) -> str:
        """
        Returns a structured key separated by / between kind and name.
        This format is suitable for expressing in a form embedded in a URL.
        """
        return '/'.join([
            f'{pair.kind}/{pair.name}' for pair in self.pairs()
        ])

    def key_path_urlsafe(self) -> str:
        return self.key_path().replace('/', '%2F').replace(':', '%3A')


class EntityKeyFactory:
    def __init__(self):
        pass

    def build_from_pairs(self, pairs: List[Union[tuple, dict]]) -> EntityKey:
        _pairs = []
        for pair in pairs:
            if isinstance(pair, dict):
                key_pair = KeyPair(pair.get('kind'), pair.get('name'))
            elif isinstance(pair, tuple) or isinstance(pair, list):
                key_pair = KeyPair(pair[0], pair[1])
            else:
                raise ValueError(f'Unknown object type: {type(pair)}, got: {pair}')
            _pairs.append(key_pair)

        return EntityKey(_pairs)

    def build(self, kind: str, name: str, parent: Optional[EntityKey] = None) -> EntityKey:
        if parent:
            pairs = copy.deepcopy(parent.pairs())
        else:
            pairs = []
        pairs.append(KeyPair(kind=kind, name=name))

        return EntityKey(pairs)

    def build_for_new(self, kind: str, parent: Optional[EntityKey] = None) -> EntityKey:
        name = self._generate_new_uuid()
        return self.build(kind=kind, name=name, parent=parent)

    def _generate_new_uuid(self) -> str:
        s = base64.b32encode(uuid.uuid4().bytes).decode('utf-8')
        return s.replace('======', '').lower()

    def build_from_key_path(self, key_path: str) -> EntityKey:
        pairs = []
        for pair in key_path.replace('%2F', '/').replace('%3A', ':').split('/'):
            kind, name = pair.split(':')
            pairs.append(KeyPair(kind=kind, name=name))

        return EntityKey(pairs)

    @staticmethod
    def _split_list(l, n):
        """
        list divided by n items.
        :param l: list
        :param n: item size of sub-list
        :return:
        """
        for idx in range(0, len(l), n):
            yield l[idx:idx + n]

    def build_from_key_url(self, key_url: str) -> EntityKey:
        pairs = []

        key_elements = key_url.replace('%2F', '/').split('/')
        key_pairs = list(self._split_list(key_elements, 2))

        for pair in key_pairs:
            pairs.append(KeyPair(kind=pair[0], name=pair[1]))

        return EntityKey(pairs)
